"""Top-level package for smunger."""

import logging

from rich.logging import RichHandler

from .console import console
from .constant import ColAllowNA, ColName, ColRange, ColType
from .io import load_sumstats, save_sumstats
from .mapheader import map_colnames
from .smunger import (
    make_SNPID_unique,
    extract_cols,
    munge,
    munge_allele,
    munge_beta,
    munge_bp,
    munge_chr,
    munge_eaf,
    munge_maf,
    munge_or,
    munge_pvalue,
    munge_se,
    munge_z,
)

__author__ = """Jianhua Wang"""
__email__ = 'jianhua.mert@gmail.com'
__version__ = '0.0.6'

# Set up logging
logging.basicConfig(
    level="INFO",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)],
)



"""Main module."""

import json
from typing import Optional
from pathlib import Path

import pandas as pd
from rich.prompt import Confirm
from rich.table import Table

from smunger.console import console
from smunger.constant import COMMON_COLNAMES, ColName

# display first five rows of input dataframe, and guess the column names
# that are most likely to be the column names of munged dataframe
# display the guessed column map, let user to confirm
# if user confirms, then use the guessed column map
# if user does not confirm, then ask user to input the column map
# batch mode: remember the column map for each input file


def display_df(df: pd.DataFrame, colname_map: dict, nrows: int = 5):
    """Display the first five rows of a dataframe."""
    if len(df) < 5:
        nrows = len(df)
    console.print(f"Displaying first {nrows} rows of the dataframe:")
    console.print("First row is the number of the column.")
    console.print("Second row is the original column name.")
    console.print("Third row is the gussed mapped column name.")
    table = Table(show_header=True, header_style="bold magenta")
    original_colnames = []
    mapped_colnames = []
    value_colidx = []
    for i, col in enumerate(df.columns):
        table.add_column(f"No.{i}", justify="center", style="cyan", no_wrap=False)

        original_colnames.append(col)
        value_colidx.append(i)
        if col in colname_map:
            mapped_colnames.append(colname_map[col])
        else:
            mapped_colnames.append('')
        if (i + 1) % 10 == 0 or i == len(df.columns) - 1:
            table.add_row(*original_colnames, style="bold red", end_section=True)
            table.add_row(*mapped_colnames, style="bold green", end_section=True)
            for n in range(nrows):
                table.add_row(*[str(_) for _ in df.iloc[n, value_colidx].values])
            console.print(table)
            table = Table(show_header=True, header_style="bold magenta")
            original_colnames = []
            mapped_colnames = []
            value_colidx = []
    missing_colnames = set(ColName.OUTCOLS) - set(colname_map.values())
    if missing_colnames:
        console.print(f"Missing columns: {missing_colnames}")


def map_colnames(df: pd.DataFrame, outfile: Optional[Path] = None) -> dict:
    """Map column names."""
    # map column names
    guessed_map = guess_colnames(df)
    display_df(df, guessed_map)
    use_guessed_map = Confirm.ask("Do you want to use the guessed column map?", default=True)
    if use_guessed_map:
        colname_map = guessed_map
    else:
        colname_map = {}
        guess_orignal_out = {v: k for k, v in guessed_map.items()}
        for col in ColName.OUTCOLS:
            if col in guess_orignal_out:
                map_correct = Confirm.ask(
                    f"Is {guess_orignal_out[col]} the correct column name for {col}?", default=True
                )
                if map_correct:
                    colname_map[guess_orignal_out[col]] = col
                    continue
            manual_map = input(f"Please input the correct column number for {col}: ")
            if manual_map == '':
                continue
            else:
                manual_map = df.columns[int(manual_map)]
                colname_map[manual_map] = col
        # use OR if BETA is not found
        if ColName.BETA not in colname_map.values() or ColName.SE not in colname_map.values():
            console.print("BETA is not found.")
            for col in [ColName.OR, ColName.ORSE, ColName.Z]:
                if col in guess_orignal_out:
                    map_correct = Confirm.ask(
                        f"Is {guess_orignal_out[col]} the correct column name for {col}?", default=True
                    )
                    if map_correct:
                        colname_map[guess_orignal_out[col]] = col
                    else:
                        manual_map = input(f"Please input the correct column number for {col}: ")
                        if manual_map == '':
                            continue
                        else:
                            manual_map = df.columns[int(manual_map)]
                            colname_map[manual_map] = col
    console.print(f"Column map is: {colname_map}")
    display_df(df, colname_map)
    if outfile:
        console.print(f"Saving column map to {outfile}")
        with open(outfile, 'w') as f:
            json.dump(colname_map, f, indent=4)
    return colname_map


def guess_colnames(df: pd.DataFrame) -> dict:
    """Guess column names."""
    # guess column names
    colnames = {}
    for col in df.columns:
        if col in COMMON_COLNAMES:
            colnames[col] = COMMON_COLNAMES[col]
    return colnames



"""Console script for smunger."""

import logging
from pathlib import Path

import typer

from smunger import __version__, console

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
app = typer.Typer(context_settings=CONTEXT_SETTINGS, add_completion=False)


@app.callback(invoke_without_command=True, no_args_is_help=True)
def main(
    version: bool = typer.Option(False, '--version', '-V', help='Show version.'),
    verbose: bool = typer.Option(False, '--verbose', '-v', help='Show verbose info.'),
):
    """smunger: munger for GWAS summary statistics."""
    console.rule("[bold blue]smunger[/bold blue]")
    console.print(f"Version: {__version__}", justify="center")
    console.print("Author: Jianhua Wang", justify="center")
    console.print("Email: jianhua.mert@gmail.com", justify="center")
    if version:
        typer.echo(f'smunger version: {__version__}')
        raise typer.Exit()
    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.info('Verbose mode is on.')
    else:
        logging.getLogger().setLevel(logging.INFO)


@app.command()
def mapheader(
    infile: Path = typer.Argument(..., help='Input summary statistics.'),
    outfile: Path = typer.Argument(..., help='Output file, json.'),
    sep: str = typer.Option(None, '--sep', '-s', help='Separator of the input file.'),
    nrows: int = typer.Option(5, '--nrows', '-n', help='Number of rows to display.'),
    skiprows: int = typer.Option(0, '--skiprows', '-k', help='Number of rows to skip.'),
    comment: str = typer.Option(None, '--comment', '-c', help='Comment character.'),
    gzipped: bool = typer.Option(None, '--gzipped', '-z', help='Input file is gzipped.'),
):
    """Map column names."""
    from smunger.io import load_sumstats
    from smunger.mapheader import map_colnames
    df = load_sumstats(infile, nrows=nrows, sep=sep, skiprows=skiprows, comment=comment, gzipped=gzipped)
    map_colnames(df, outfile=outfile)


@app.command()
def munge(
    infile: Path = typer.Argument(..., help='Input summary statistics.'),
    outfile: Path = typer.Argument(..., help='Output munged summary statistics.'),
    colmap: str = typer.Argument(..., help='Column map file, json.'),
    sep: str = typer.Option(None, '--sep', '-s', help='Separator of the input file.'),
    skiprows: int = typer.Option(0, '--skiprows', '-k', help='Number of rows to skip.'),
    comment: str = typer.Option(None, '--comment', '-c', help='Comment character.'),
    gzipped: bool = typer.Option(None, '--gzipped', '-z', help='Input file is gzipped.'),
    build_index: bool = typer.Option(True, '--build-index', '-b', help='Build tabix index.'),
    sigsnps: Path = typer.Option(None, '--sigsnps', '-S', help='save significant SNPs to file.'),
    sigsnps_pval: float = typer.Option(5e-8, '--sigsnps-pval', '-P', help='p-value threshold for significant SNPs.'),
    report: str = typer.Option(None, '--report', '-R', help='save report to file.'),
):
    """Munge summary statistics."""
    import smunger
    from smunger.io import load_sumstats, save_sumstats
    df = load_sumstats(infile, sep=sep, skiprows=skiprows, comment=comment, gzipped=gzipped)
    pre_nrow = len(df)
    df = smunger.extract_cols(df, colname_map=colmap)
    df = smunger.munge(df)
    after_nrow = len(df)
    save_sumstats(df, outfile, build_index=build_index)
    from smunger.smunger import get_sigdf
    df_sig = get_sigdf(df, pval=sigsnps_pval)
    if sigsnps:
        if len(df_sig) > 0:
            save_sumstats(df_sig, sigsnps, build_index=build_index)
        else:
            console.print('[bold red]No significant SNPs found.[/bold red]')
    non_null_cols = df.columns[df.notnull().any()].tolist()
    report_json = {
        'in_rows': pre_nrow,
        'out_rows': after_nrow,
        'sigsnps': len(df_sig),
        'non_null_cols': non_null_cols,
    }
    if report:
        import json
        with open(report, 'w') as f:
            json.dump(report_json, f, indent=4)


if __name__ == "__main__":
    app()



"""Liftover summary statistics from one genome build to another."""

import logging
from typing import Tuple

import pandas as pd
import requests  # type: ignore
from liftover import get_lifter

from smunger.constant import ColName
from smunger.smunger import munge_bp, munge_chr

logger = logging.getLogger('liftover')


def liftover_singlesnp(inbuild: str, outbuild: str, chrom: int, pos: int) -> Tuple[int, int]:
    """Liftover a single SNP from one genome build to another."""
    lo = get_lifter(inbuild, outbuild)
    out = lo.query(chrom, pos)
    if len(out) > 0:
        return int(out[0][0][3:]), out[0][1]
    else:
        return 0, 0


def guess_build_by_singleSNP(rsid: str, pos: int) -> str:
    """Guess the genome build of a single SNP."""
    build_guess = ''
    try:
        a = requests.get(f'https://myvariant.info/v1/variant/{rsid}')
        a = a.json()
        chrom = a['chrom']
        hg19 = int(a['hg19']['end'])

        coordinates = {'hg19': hg19}
        for build in ['hg17', 'hg18', 'hg38']:
            coordinates[build] = liftover_singlesnp('hg19', build, chrom, hg19)[1]

        if coordinates['hg19'] == coordinates['hg17'] == coordinates['hg18'] == coordinates['hg38'] == pos:
            build_guess = ''
        elif coordinates['hg19'] == pos:
            build_guess = 'hg19'
        elif coordinates['hg17'] == pos:
            build_guess = 'hg17'
        elif coordinates['hg18'] == pos:
            build_guess = 'hg18'
        elif coordinates['hg38'] == pos:
            build_guess = 'hg38'
        else:
            build_guess = ''
    except Exception:
        pass
    finally:
        return build_guess


def guess_genome_build(df: pd.DataFrame) -> str:
    """Guess the genome build of a summary statistics file."""
    build_guess = ''
    if ColName.RSID in df.columns:
        df = df[df[ColName.RSID].notnull()]
        df = df[df[ColName.RSID].str.startswith('rs')]
        if len(df) > 0:
            guess_df = df.sample(5)
            all_guess = []
            for i in range(len(guess_df)):
                guess = guess_build_by_singleSNP(guess_df.iloc[i][ColName.RSID], guess_df.iloc[i][ColName.BP])
                if guess != '':
                    all_guess.append(guess)
            if len(all_guess) > 0:
                build_guess = max(set(all_guess), key=all_guess.count)
    return build_guess


def liftover(df: pd.DataFrame, inbuild: str, outbuild: str) -> pd.DataFrame:
    """Liftover summary statistics from one genome build to another."""
    lo = get_lifter(inbuild, outbuild)
    df[outbuild] = df[[ColName.CHR, ColName.BP]].apply(lambda x: lo.query(x[0], x[1]), axis=1)
    df[ColName.CHR] = df[outbuild].apply(lambda x: x[0][0] if len(x) > 0 else 0)
    df[ColName.BP] = df[outbuild].apply(lambda x: x[0][1] if len(x) > 0 else 0)
    df.drop(outbuild, axis=1, inplace=True)
    df = munge_chr(df)
    df = munge_bp(df)
    return df



"""Main module."""

import json
import logging
from typing import Union

import pandas as pd

from smunger.constant import ColName, ColRange, ColType

logger = logging.getLogger('munger')


def make_SNPID_unique(sumstat: pd.DataFrame) -> pd.DataFrame:
    """
    Make the SNPID unique.

    The unique SNPID is chr-bp-sorted(EA,NEA)

    Parameters
    ----------
    sumstat : pd.DataFrame
        The input summary statistics.

    Returns
    -------
    pd.DataFrame
        The summary statistics with unique SNPID.
    """
    df = sumstat.copy()
    allele_df = df[[ColName.EA, ColName.NEA]].copy()
    b = allele_df.values
    b.sort(axis=1)
    allele_df[[ColName.EA, ColName.NEA]] = b
    allele_df[ColName.SNPID] = (
        df[ColName.CHR].astype(str)
        + "-"
        + df[ColName.BP].astype(str)
        + "-"
        + allele_df[ColName.EA]
        + "-"
        + allele_df[ColName.NEA]
    )
    df.insert(loc=0, column=ColName.SNPID, value=allele_df[ColName.SNPID].values)  # type: ignore
    return df


def extract_cols(df: pd.DataFrame, colname_map: Union[dict, str]) -> pd.DataFrame:
    """Map column names."""
    # map column names
    if isinstance(colname_map, str):
        with open(colname_map, 'r') as f:
            colname_map = json.load(f)
    outdf = df.rename(columns=colname_map).copy()  # type: ignore
    mapped_cols = list(colname_map.values())  # type: ignore
    outdf = outdf[mapped_cols]
    outdf = rm_col_allna(outdf)
    return outdf


def checl_colnames(df: pd.DataFrame) -> pd.DataFrame:
    """Check column names, fill None if not presents."""
    outdf = df.copy()
    for col in ColName.OUTCOLS:
        if col not in outdf.columns:
            outdf[col] = None
    return outdf[ColName.OUTCOLS]


def rm_col_allna(df: pd.DataFrame) -> pd.DataFrame:
    """Remove columns that are all NA."""
    outdf = df.copy()
    for col in outdf.columns:
        if outdf[col].isnull().all():
            logger.debug(f"Remove column {col} because it is all NA.")
            outdf.drop(col, axis=1, inplace=True)
    return outdf


def get_sigdf(df: pd.DataFrame, pval: float = 5e-8) -> pd.DataFrame:
    """Get significant SNPs."""
    outdf = df.copy()
    if ColName.P in outdf.columns:
        outdf = outdf[outdf[ColName.P] < pval]
    else:
        raise ValueError("Missing P column.")
    return outdf


def munge(df: pd.DataFrame) -> pd.DataFrame:
    """Munge the summary statistics."""
    outdf = df.copy()
    if all(
        [
            ColName.CHR in outdf.columns,
            ColName.BP in outdf.columns,
            ColName.EA in outdf.columns,
            ColName.NEA in outdf.columns,
        ]
    ):
        outdf = munge_chr(outdf)
        outdf = munge_bp(outdf)
        outdf = munge_allele(outdf)
        outdf = make_SNPID_unique(outdf)
        if ColName.P in outdf.columns:
            outdf = munge_pvalue(outdf)
            outdf = outdf.sort_values(by=ColName.P)
        else:
            pass
        pre_n = outdf.shape[0]
        outdf = outdf.drop_duplicates(subset=ColName.SNPID, keep="first")
        outdf = outdf.sort_values(by=[ColName.CHR, ColName.BP])
        after_n = outdf.shape[0]
        logger.debug(f"Remove {pre_n - after_n} duplicated SNPs.")
    else:
        raise ValueError("Missing CHR, BP, EA or NEA column.")

    # outdf = munge_rsid(outdf)
    if ColName.BETA in outdf.columns and ColName.SE in outdf.columns:
        outdf = munge_beta(outdf)
        outdf = munge_se(outdf)
    elif ColName.OR in outdf.columns and ColName.ORSE in outdf.columns:
        outdf = munge_or(outdf)
        outdf = munge_orse(outdf)
        outdf[ColName.BETA] = outdf[ColName.OR]
        outdf[ColName.SE] = outdf[ColName.ORSE] / outdf[ColName.OR]
        outdf = munge_beta(outdf)
        outdf = munge_se(outdf)
        del outdf[ColName.OR]
        del outdf[ColName.ORSE]
    else:
        logger.warning("Missing BETA or SE column.")

    if ColName.Z in outdf.columns:
        outdf = munge_z(outdf)

    if ColName.EAF in outdf.columns:
        outdf = munge_eaf(outdf)
        outdf[ColName.MAF] = outdf[ColName.EAF]
    if ColName.MAF in outdf.columns:
        outdf = munge_maf(outdf)
    outdf = checl_colnames(outdf)
    return outdf


def munge_rsid(df: pd.DataFrame) -> pd.DataFrame:
    """Munge rsID column."""
    outdf = df.copy()
    outdf[ColName.RSID] = outdf[ColName.RSID].astype(ColType.RSID)
    return outdf


def munge_chr(df: pd.DataFrame) -> pd.DataFrame:
    """Munge chromosome column."""
    pre_n = df.shape[0]
    outdf = df[df[ColName.CHR].notnull()].copy()
    outdf[ColName.CHR] = outdf[ColName.CHR].astype(str)
    outdf[ColName.CHR] = outdf[ColName.CHR].str.replace("chr", "")
    # replace X, with 23
    outdf[ColName.CHR] = outdf[ColName.CHR].replace("X", 23)
    outdf[ColName.CHR] = outdf[ColName.CHR].replace("x", 23)
    # turn chromosome column into integer
    outdf[ColName.CHR] = pd.to_numeric(outdf[ColName.CHR], errors="coerce")
    outdf = outdf[outdf[ColName.CHR].notnull()]
    outdf = outdf[(outdf[ColName.CHR] >= ColRange.CHR_MIN) & (outdf[ColName.CHR] <= ColRange.CHR_MAX)]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid chromosome.")
    outdf[ColName.CHR] = outdf[ColName.CHR].astype(ColType.CHR)
    return outdf


def munge_bp(df: pd.DataFrame) -> pd.DataFrame:
    """Munge position column."""
    pre_n = df.shape[0]
    outdf = df[df[ColName.BP].notnull()].copy()
    outdf[ColName.BP] = pd.to_numeric(outdf[ColName.BP], errors="coerce")
    outdf = outdf[outdf[ColName.BP].notnull()]
    outdf = outdf[(outdf[ColName.BP] > ColRange.BP_MIN) & (outdf[ColName.BP] < ColRange.BP_MAX)]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid position.")
    outdf[ColName.BP] = outdf[ColName.BP].astype(ColType.BP)
    return outdf


def munge_allele(df: pd.DataFrame) -> pd.DataFrame:
    """Munge allele column."""
    outdf = df.copy()
    for col in [ColName.EA, ColName.NEA]:
        pre_n = outdf.shape[0]
        outdf = outdf[outdf[col].notnull()]
        outdf[col] = outdf[col].astype(str)
        outdf[col] = outdf[col].str.upper()
        # make sure all alleles only contain one or more ACGT characters
        outdf = outdf[outdf[col].str.match(r"^[ACGT]+$")]
        after_n = outdf.shape[0]
        logger.debug(f"Remove {pre_n - after_n} rows because of invalid {col}.")
    outdf = outdf[outdf[ColName.EA] != outdf[ColName.NEA]]
    return outdf


def munge_pvalue(df: pd.DataFrame) -> pd.DataFrame:
    """Munge pvalue column."""
    outdf = df.copy()
    pre_n = outdf.shape[0]
    outdf[ColName.P] = pd.to_numeric(outdf[ColName.P], errors="coerce")
    outdf = outdf[outdf[ColName.P].notnull()]
    outdf = outdf[(outdf[ColName.P] > ColRange.P_MIN) & (outdf[ColName.P] < ColRange.P_MAX)]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid pvalue.")
    outdf[ColName.P] = outdf[ColName.P].astype(ColType.P)
    return outdf


def munge_beta(df: pd.DataFrame) -> pd.DataFrame:
    """Munge beta column."""
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.BETA] = pd.to_numeric(outdf[ColName.BETA], errors="coerce")
    outdf = outdf[outdf[ColName.BETA].notnull()]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid beta.")
    outdf[ColName.BETA] = outdf[ColName.BETA].astype(ColType.BETA)
    return outdf


def munge_se(df: pd.DataFrame) -> pd.DataFrame:
    """Munge se column."""
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.SE] = pd.to_numeric(outdf[ColName.SE], errors="coerce")
    outdf = outdf[outdf[ColName.SE].notnull()]
    outdf = outdf[outdf[ColName.SE] > ColRange.SE_MIN]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid se.")
    outdf[ColName.SE] = outdf[ColName.SE].astype(ColType.SE)
    return outdf


def munge_or(df: pd.DataFrame) -> pd.DataFrame:
    """Munge or column."""
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.OR] = pd.to_numeric(outdf[ColName.OR], errors="coerce")
    outdf = outdf[outdf[ColName.OR].notnull()]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid or.")
    outdf[ColName.OR] = outdf[ColName.OR].astype(ColType.OR)
    return outdf


def munge_orse(df: pd.DataFrame) -> pd.DataFrame:
    """Munge orse column."""
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.ORSE] = pd.to_numeric(outdf[ColName.ORSE], errors="coerce")
    outdf = outdf[outdf[ColName.ORSE].notnull()]
    outdf = outdf[outdf[ColName.ORSE] > ColRange.ORSE_MIN]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid orse.")
    outdf[ColName.ORSE] = outdf[ColName.ORSE].astype(ColType.ORSE)
    return outdf


def munge_z(df: pd.DataFrame) -> pd.DataFrame:
    """Munge z column."""
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.Z] = pd.to_numeric(outdf[ColName.Z], errors="coerce")
    outdf = outdf[outdf[ColName.Z].notnull()]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid z.")
    outdf[ColName.Z] = outdf[ColName.Z].astype(ColType.Z)
    return outdf


def munge_eaf(df: pd.DataFrame) -> pd.DataFrame:
    """Munge eaf column."""
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.EAF] = pd.to_numeric(outdf[ColName.EAF], errors="coerce")
    outdf = outdf[outdf[ColName.EAF].notnull()]
    outdf = outdf[(outdf[ColName.EAF] >= ColRange.EAF_MIN) & (outdf[ColName.EAF] <= ColRange.EAF_MAX)]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid eaf.")
    outdf[ColName.EAF] = outdf[ColName.EAF].astype(ColType.EAF)
    return outdf


def munge_maf(df: pd.DataFrame) -> pd.DataFrame:
    """Munge maf column."""
    pre_n = df.shape[0]
    outdf = df.copy()
    outdf[ColName.MAF] = pd.to_numeric(outdf[ColName.MAF], errors="coerce")
    outdf = outdf[outdf[ColName.MAF].notnull()]
    outdf[ColName.MAF] = outdf[ColName.MAF].apply(lambda x: 1 - x if x > 0.5 else x)
    outdf = outdf[(outdf[ColName.MAF] >= ColRange.MAF_MIN) & (outdf[ColName.MAF] <= ColRange.MAF_MAX)]
    after_n = outdf.shape[0]
    logger.debug(f"Remove {pre_n - after_n} rows because of invalid maf.")
    outdf[ColName.MAF] = outdf[ColName.MAF].astype(ColType.MAF)
    return outdf



"""Read and write data from/to files."""

import gzip
import logging
import shutil
from pathlib import Path
from subprocess import PIPE, run
from typing import Optional

import pandas as pd

from .constant import ColName

logger = logging.getLogger('io')


def load_sumstats(
    filename: Path,
    sep: Optional[str] = None,
    nrows: Optional[int] = None,
    skiprows: int = 0,
    comment: Optional[str] = None,
    gzipped: Optional[bool] = None,
) -> pd.DataFrame:
    """Load summary statistics from a file."""
    # determine whether the file is gzipped
    if gzipped is None:
        gzipped = filename.suffix.endswith('gz')

    # read the first line of the file to determine the separator
    if sep is None:
        if gzipped:
            with gzip.open(filename, 'rt') as f:
                line = f.readline()
        else:
            with open(filename, 'rt') as f:
                line = f.readline()
        if '\t' in line:
            sep = '\t'
        elif ',' in line:
            sep = ','
        else:
            sep = ' '

    # determine the separator, automatically if not specified
    return pd.read_csv(
        filename, sep=sep, nrows=nrows, skiprows=skiprows, comment=comment, compression='gzip' if gzipped else None
    )


def check_tool(tool: str) -> str:
    """
    Check if tool is installed.

    Parameters
    ----------
    tool : str
        The tool name.

    Returns
    -------
    str
        The path of the tool if it is installed, otherwise exit the program.
    """
    tool_path = shutil.which(tool)
    if tool_path:
        return tool_path
    else:
        logger.error(f"{tool} is not installed. Please install it first and make sure it is in your PATH.")
        raise ValueError(f"{tool} is not installed. Please install it first and make sure it is in your PATH.")


def save_sumstats(sumstats: pd.DataFrame, filename: Path, build_index: bool = True):
    """Save summary statistics to a file."""
    # save the summary statistics to a file
    if isinstance(filename, str):
        filename = Path(filename)
    if filename.exists():
        logger.warning(f'File {filename} already exists. Overwriting.')

    if filename.suffix == '.gz':
        filename = filename.with_suffix('')
    sumstats = sumstats.sort_values(by=[ColName.CHR, ColName.BP])
    sumstats.to_csv(filename, sep='\t', index=False, header=True)

    # compress the file
    bgzip = check_tool('bgzip')
    run([bgzip, '-f', filename], stdout=PIPE, stderr=PIPE, check=True)
    # index the file
    if build_index:
        tabix = check_tool('tabix')
        run(
            [tabix, '-f', '-S', '1', '-s', '1', '-b', '2', '-e', '2', f'{filename}.gz'],
            stdout=PIPE,
            stderr=PIPE,
            check=True,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smunger', 'tests']

package_data = \
{'': ['*'], 'tests': ['exampledata/*']}

install_requires = \
['liftover>=1.1.16,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'requests>=2.28.2,<3.0.0',
 'rich>=13.3.1,<14.0.0',
 'typer>=0.7.0,<0.8.0']

extras_require = \
{'dev': ['tox>=3.20.1,<4.0.0',
         'virtualenv>=20.2.2,<21.0.0',
         'pip>=20.3.1,<21.0.0',
         'twine>=3.3.0,<4.0.0',
         'pre-commit>=2.12.0,<3.0.0',
         'toml>=0.10.2,<0.11.0',
         'bump2version>=1.0.1,<2.0.0'],
 'doc': ['mkdocs>=1.4.2,<2.0.0',
         'mkdocs-include-markdown-plugin>=4.0.3,<5.0.0',
         'mkdocs-material>=8.5.11,<9.0.0',
         'mkdocs-autorefs>=0.4.1,<0.5.0',
         'mkdocstrings[python]>=0.19.1,<0.20.0'],
 'test': ['black>=22.3.0',
          'isort>=5.8.0,<6.0.0',
          'flake8>=3.9.2,<4.0.0',
          'flake8-docstrings>=1.6.0,<2.0.0',
          'mypy>=0.900,<0.901',
          'pytest>=6.2.4,<7.0.0',
          'pytest-cov>=2.12.0,<3.0.0']}

entry_points = \
{'console_scripts': ['smunger = smunger.cli:app']}

setup_kwargs = {
    'name': 'smunger',
    'version': '0.0.6',
    'description': 'munger for GWAS summary statistics.',
    'long_description': '# smunger\n\n\n[![pypi](https://img.shields.io/pypi/v/smunger.svg)](https://pypi.org/project/smunger/)\n[![python](https://img.shields.io/pypi/pyversions/smunger.svg)](https://pypi.org/project/smunger/)\n[![Build Status](https://github.com/jianhua/smunger/actions/workflows/dev.yml/badge.svg)](https://github.com/jianhua/smunger/actions/workflows/dev.yml)\n[![codecov](https://codecov.io/gh/jianhua/smunger/branch/main/graphs/badge.svg)](https://codecov.io/github/jianhua/smunger)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n\n\nmunger for GWAS summary statistics\n\n\n* Documentation: <https://jianhua.github.io/smunger>\n* GitHub: <https://github.com/jianhua/smunger>\n* PyPI: <https://pypi.org/project/smunger/>\n* Free software: MIT\n\n\n## Features\n\n- [x]  define column properties\n    - [x]  required columns: CHR, BP, EA, NEA\n    - [x]  optional columns: BETA, SE, P, EAF, MAF\n    - [x]  Auxiliary columns: OR, OR_SE, Z\n    - [x]  Data types\n    - [x]  Data ranges\n    - [x]  Allow missing values and default missing values\n- [x]  semi-automatically header mapping\n    - [x]  read first five rows and display in terminal\n    - [x]  guess header map by common column names\n    - [x]  manually check if the mapping is correct\n    - [x]  input the right column number if it is wrong\n    - [x]  check if OR, OR_SE, Z are present if BETA, SE are absent\n    - [x]  save the final column map to json for further munging\n- [x]  data munging\n    - [x]  EA ≠ NEA\n    - [x]  if EAF presents, MAF = min(EAF, 1-EAF)\n    - [x]  convert OR/ORSE to BETA/SE, if BETA, SE are absent and OR, ORSE are present\n    - [x]  remove duplicate SNPs with same chr-bp-sorted(EA,NEA), keep the one with lowest P\n    - [x]  output: \\t separated, `bgzip` compress, `tabix` index.\n    - [x]  optional output: significant SNPs, munge report\n    \n    |  | CHR | BP | rsID | EA | NEA | EAF | MAF | BETA | SE | P | OR | OR_SE | Z |\n    | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- | --- |\n    | type | int | int | str | str | str | float | float | float | float | float | float | float | float |\n    | allow null | False | False | True | False | False | False | False | True | False | True | True | False | True |\n    | null value |  |  |  |  |  |  |  | 0 |  | 0.999 | 1 |  | 0 |\n    | range | [1，23] | (0,inf) |  | only contains ‘ACGT’ | only contains ‘ACGT’ | [0,1] | [0,0.5] | (-inf,inf) | (0, inf) | (0,1) | (0, inf) | (0, inf) | (-inf,inf) |\n- [x]  liftover\n    - [x]  guess genome build\n    - [x]  liftover\n- [ ]  annotate\n    - [ ]  annotate rsID\n\n## Credits\n\nThis package was created with [Cookiecutter](https://github.com/audreyr/cookiecutter) and the [waynerv/cookiecutter-pypackage](https://github.com/waynerv/cookiecutter-pypackage) project template.\n',
    'author': 'Jianhua Wang',
    'author_email': 'jianhua.mert@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jianhua/smunger',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



"""Define the constants used in the package."""

import numpy as np


class ColName:
    """Define column names."""

    CHR = "CHR"
    BP = "BP"
    RSID = "rsID"
    SNPID = 'SNPID'  # unique snpid, chr-bp-sorted(EA,NEA)
    EA = "EA"
    NEA = "NEA"
    P = "P"
    BETA = "BETA"
    OR = "OR"
    ORSE = "ORSE"
    SE = "SE"
    EAF = "EAF"
    MAF = "MAF"
    N = "N"
    Z = "Z"
    INFO = "INFO"
    OUTCOLS = ['CHR', 'BP', 'rsID', 'EA', 'NEA', 'EAF', 'MAF', 'BETA', 'SE', 'P']


class ColType:
    """Define column types."""

    CHR = int
    BP = int
    RSID = str
    EA = str
    NEA = str
    P = float
    BETA = float
    OR = float
    ORSE = float
    SE = float
    EAF = float
    MAF = float
    N = int
    Z = float
    INFO = float


class ColAllowNA:
    """Define whether a column allows missing values."""

    CHR = False
    BP = False
    RSID = True
    EA = False
    NEA = False
    P = True
    BETA = True
    OR = True
    ORSE = True
    SE = True
    EAF = True
    MAF = True
    N = True
    Z = True
    INFO = True


class ColRange:
    """Define the range of values for each column."""

    INFO = (0, 1)
    CHR_MIN = 1
    CHR_MAX = 23
    BP_MIN = 0
    BP_MAX = 300000000
    P_MIN = 0
    P_MAX = 1
    SE_MIN = 0
    SE_MAX = np.inf
    EAF_MIN = 0
    EAF_MAX = 1
    MAF_MIN = 0
    MAF_MAX = 1
    ORSE_MIN = 0
    ORSE_MAX = np.inf


COMMON_COLNAMES = {
    'CHR': 'CHR',
    'chr': 'CHR',
    'chromosome': 'CHR',
    '#CHROM': 'CHR',
    '#CHR': 'CHR',
    'Chromosome': 'CHR',
    'CHROM': 'CHR',
    'Chr': 'CHR',
    'BP': 'BP',
    'position': 'BP',
    'pos': 'BP',
    'bp': 'BP',
    'base_pair_location': 'BP',
    'POS': 'BP',
    'Position': 'BP',
    'Position_b37': 'BP',
    'Bp': 'BP',
    'MarkerName': 'rsID',
    'varID': 'rsID',
    'rsid': 'rsID',
    'snp': 'rsID',
    'SNP': 'rsID',
    'hm_rsid': 'rsID',
    'variant_id': 'rsID',
    'uniqID': 'rsID',
    'ID': 'rsID',
    'id': 'rsID',
    'rsID': 'rsID',
    'Variant_ID': 'rsID',
    'Test': 'rsID',
    '#SNP': 'rsID',
    'MAF': 'MAF',
    'all_maf': 'MAF',
    'beta': 'BETA',
    'BETA_INSOMNIA': 'BETA',
    'BETA': 'BETA',
    'all_inv_var_meta_beta': 'BETA',
    'Effect': 'BETA',
    'lnOR': 'BETA',
    'Beta': 'BETA',
    'LogOR': 'BETA',
    'frequentist_add_beta_1': 'BETA',
    'se': 'SE',
    'or_se': 'SE',
    'standard_error': 'SE',
    'SE_INSOMNIA': 'SE',
    'SE': 'SE',
    'sebeta': 'SE',
    'all_inv_var_meta_sebeta': 'SE',
    'StdErr': 'SE',
    'Standard_error': 'SE',
    'SE_GC': 'SE',
    'StdErrLogOR': 'SE',
    'frequentist_add_se_1': 'SE',
    'P-value': 'P',
    'Score.pval': 'P',
    'p': 'P',
    'p.metal': 'P',
    'P': 'P',
    'p_value': 'P',
    'P_INSOMNIA': 'P',
    'P_BOLT_LMM_INF': 'P',
    'p.value': 'P',
    'pval': 'P',
    'all_inv_var_meta_p': 'P',
    'PS': 'P',
    'P_Value': 'P',
    'Pvalue': 'P',
    'P_GC': 'P',
    'P.value': 'P',
    'Zscore': 'Zscore',
    'Z': 'Zscore',
    'z_value': 'Zscore',
    'z.meta': 'Zscore',
    'Z-score': 'Zscore',
    'Freq1': 'EAF',
    'altFreq1': 'EAF',
    'af': 'EAF',
    'freqa1': 'EAF',
    'effect_allele_frequency': 'EAF',
    'A1FREQ': 'EAF',
    'all_meta_AF': 'EAF',
    'FREQ_Effect_Allele': 'EAF',
    'EAF': 'EAF',
    'Frq': 'EAF',
    'Coded_freq': 'EAF',
    'Freq': 'EAF',
    'eaf': 'EAF',
    'or': 'OR',
    'odds_ratio': 'OR',
    'OR': 'OR',
    'ORX': 'OR',
    'ORS': 'OR',
    'Odds_ratio': 'OR',
    'other_allele': 'NEA',
    'effect_allele': 'EA',
    'Allele1': 'EA',
    'Allele2': 'NEA',
    'Freq.Allele1.HapMapCEU': 'EAF',
    'allele1': 'EA',
    'allele2': 'NEA',
    'freqA1': 'EAF',
    'A1': 'EA',
    'A2': 'NEA',
    'maf': 'MAF',
    'beta_SNP_add': 'BETA',
    'sebeta_SNP_add': 'SE',
    'Minor_allele': 'EA',
    'all_OR': 'OR',
    'frequentist_add_pvalue': 'P',
    'z': 'Zscore',
    'SNP_ID': 'rsID',
    'Allele 1': 'EA',
    'Allele 2': 'NEA',
    'Effect allele (EA)': 'EA',
    'Effect allele frequency (EAF)': 'EAF',
    'P value': 'P',
    'AF': 'EAF',
    'ALLELE1': 'EA',
    'ALLELE0': 'NEA',
    'P_LINREG': 'P',
    'snpid': 'rsID',
    'bpos': 'BP',
    'a1': 'EA',
    'a2': 'NEA',
    'mtag_beta': 'BETA',
    'mtag_se': 'SE',
    'mtag_z': 'Zscore',
    'mtag_pval': 'P',
    'position_GRCh37': 'BP',
    'non_effect_allele': 'NEA',
    'REF': 'NEA',
    'ALT': 'EA',
    'noneffect_allele': 'NEA',
    'FreqSE': 'SE',
    'ALT_ALLELE_FREQ': 'EAF',
    'PVALUE': 'Zscore',
    'EAF_A1': 'EAF',
    'Pval': 'P',
    'RSID': 'rsID',
    'a0': 'NEA',
    'freq1': 'EAF',
    'freq_se': 'SE',
    'EA': 'EA',
    'NEA': 'NEA',
    'LOGOR': 'BETA',
    'LOGOR_SE': 'SE',
    'Effect_allele': 'EA',
    'Non_Effect_allele': 'NEA',
    'PVAL': 'P',
    'Effect_Allele': 'EA',
    'NonEffect_Allele': 'NEA',
    'Effect_Allele_Frequency': 'EAF',
    'SE_of_Beta': 'SE',
    'A1Frq': 'EAF',
    'FREQ': 'EAF',
    'a_0': 'NEA',
    'a_1': 'EA',
    'pvalue': 'P',
    'Pos_b37': 'BP',
    'Allele_1': 'EA',
    'Allele_2': 'NEA',
    'Other_allele': 'NEA',
    'Effect_allele_Freq': 'EAF',
    'markername': 'rsID',
    'effect_allele_freq': 'EAF',
    'EFFECT': 'BETA',
    'STDERR': 'SE',
    'OA': 'NEA',
    'BPos': 'BP',
    'Non_effect_Allele': 'NEA',
    'Standard_Error_of_Beta': 'SE',
    'p-value': 'P',
    'snp_ids': 'rsID',
    'Ref_allele': 'NEA',
    'Alt_allele': 'EA',
    'JASS_PVAL': 'P',
    'variant_ID': 'rsID',
    'alleleA': 'EA',
    'alleleB': 'NEA',
    'A1Freq': 'EAF',
    'MAF_calculated_from_dosage_data': 'MAF',
    'REF_allele': 'NEA',
    'ALT_allele': 'EA',
    'none_effect_alllele': 'NEA',
    'Pos': 'BP',
    'ZSCORE': 'Zscore',
    'zscore': 'Zscore',
    'allele_1': 'EA',
    'allele_2': 'NEA',
    'allele0': 'NEA',
    'snp.1': 'rsID',
    'a1freq': 'EAF',
    'snpname': 'rsID',
    'ref': 'NEA',
    'alt': 'EA',
    'A0': 'NEA',
    'reference_allele': 'NEA',
    'base_pair_location_grch37': 'BP',
    'Rsid': 'rsID',
    'Non_effect_allele': 'NEA',
    'EffectAllele': 'EA',
    'NonEffectAllele': 'NEA',
    'Freq.A1.ESP.EUR': 'EAF',
    'base_pair_position': 'BP',
    'effect_allle_frequency': 'EAF',
    'base_pair_locations': 'BP',
    'ZScore': 'Zscore',
    'stderr': 'SE',
}
